import builtins
import json
import random
import time

from seahorse.game.io_stream import EventSlave


class StateRecorder(EventSlave):
    """ An event slave that records everything state emitted by the master in a json file
    """

    def __init__(self) -> None:
        super().__init__() 
        self.identifier = "__REC__"+str(int(time.time()*1000000-random.randint(1,1000000)))
        self.id = builtins.id(self)
        self.wrapped_id = self.id
        self.sid = None

        self.activate(self.identifier)
        self.recorded_content = []

        @self.sio.on("play")
        def record_play(data):
            self.recorded_content.append(json.loads(data)) 

        @self.sio.event()
        def disconnect():
            with open(self.identifier+".json","w+") as f:
                f.write(json.dumps(self.recorded_content))





import builtins
import os
import subprocess
from typing import Any, Coroutine, Optional

from loguru import logger
from seahorse.game.io_stream import EventMaster, EventSlave 

class GUIClient(EventSlave):
    def __init__(self, path:Optional[str]=None) -> None:
        self.id = builtins.id(self)
        self.wrapped_id = self.id
        self.path = path
        self.sid = None

    @staticmethod
    def open_file(url):
        try:
            os.startfile(url)
        except AttributeError:
            try:
                subprocess.call(["open", url])
            except Exception:
                logger.debug("Could not open URL")

    async def listen(self,**_) -> Coroutine[Any, Any, None]:
        if self.path:
            GUIClient.open_file(self.path)
        idmap = await EventMaster.get_instance().wait_for_identified_client("__GUI__",self.id)
        self.sid = idmap["sid"]



import copy
import json
import sys
import time
from abc import abstractmethod
from collections.abc import Iterable
from itertools import cycle
from typing import List, Optional

from loguru import logger

from seahorse.game.game_state import GameState
from seahorse.game.io_stream import EventMaster, EventSlave
from seahorse.player.player import Player
from seahorse.utils.custom_exceptions import (
    ActionNotPermittedError,
    MethodNotImplementedError,
    PlayerDuplicateError,
    SeahorseTimeoutError,
    StopAndStartError,
)


class GameMaster:
    """
    A class representing the game master.

    Attributes:
        name (str): The name of the game.
        initial_game_state (GameState): The initial state of the game.
        current_game_state (GameState): The current state of the game.
        players_iterator (Iterable): An iterable for the players, ordered according
            to the playing order. If a list is provided, a cyclic iterator is automatically built.
        log_level (str): The name of the log file.
    """

    def __init__(
        self,
        name: str,
        initial_game_state: GameState,
        players_iterator: Iterable[Player],
        log_level: str = "INFO",
        port: int =8080,
        hostname: str ="localhost"
    ) -> None:
        """
        Initializes a new instance of the GameMaster class.

        Args:
            name (str): The name of the game.
            initial_game_state (GameState): The initial state of the game.
            players_iterator (Iterable[Player]): An iterable for the players, ordered according
                to the playing order.
            log_level (str): The name of the log file.
        """
        self.timetol = 1e-1
        self.name = name
        self.current_game_state = initial_game_state
        self.players = initial_game_state.players
        player_names = [x.name for x in self.players]
        if len(set(player_names))<len(self.players):
            logger.error("Multiple players have the same name this is not allowed.")
            logger.error("Please rename your players such that there is no duplicate in the following list: ")
            logger.error(f"{player_names}")
            raise PlayerDuplicateError()


        self.log_level = log_level
        self.players_iterator = cycle(players_iterator) if isinstance(players_iterator, list) else players_iterator
        next(self.players_iterator)
        self.emitter = EventMaster.get_instance(initial_game_state.__class__,port=port,hostname=hostname)
        logger.remove()
        logger.add(sys.stderr, level=log_level)

    async def step(self) -> GameState:
        """
        Calls the next player move.

        Returns:
            GamseState : The new game_state.
        """
        next_player = self.current_game_state.get_next_player()
        possible_actions = self.current_game_state.get_possible_actions()

        start = time.time()
        next_player.start_timer()
        logger.info(f"time : {next_player.get_remaining_time()}")

        if isinstance(next_player,EventSlave):
            action = await next_player.play(self.current_game_state)
        else:
            action = next_player.play(self.current_game_state)

        tstp = time.time()
        if abs((tstp-start)-(tstp-next_player.get_last_timestamp()))>self.timetol:
            next_player.stop_timer()
            raise StopAndStartError()

        next_player.stop_timer()

        if action not in possible_actions:
            raise ActionNotPermittedError()

        action.current_game_state._possible_actions=None
        action.current_game_state=None
        return action.get_next_game_state()

    async def play_game(self) -> list[Player]:
        """
        Play the game.

        Returns:
            Iterable[Player]: The winner(s) of the game.
        """
        await self.emitter.sio.emit(
            "play",
            json.dumps(self.current_game_state.to_json(),default=lambda x:x.to_json()),
        )
        for player in self.get_game_state().get_players() :
            logger.info(f"Player : {player.get_name()} - {player.get_id()}")
        while not self.current_game_state.is_done():
            try:
                logger.info(f"Player now playing : {self.get_game_state().get_next_player().get_name()} - {self.get_game_state().get_next_player().get_id()}")
                self.current_game_state = await self.step()
            except (ActionNotPermittedError,SeahorseTimeoutError,StopAndStartError) as e:
                if isinstance(e,SeahorseTimeoutError):
                    logger.error(f"Time credit expired for player {self.current_game_state.get_next_player()}")
                elif isinstance(e,ActionNotPermittedError) :
                    logger.error(f"Action not permitted for player {self.current_game_state.get_next_player()}")
                else:
                    logger.error(f"Player {self.current_game_state.get_next_player()} might have tried tampering with the timer.\n The timedelta difference exceeded the allowed tolerancy in GameMaster.timetol ")

                temp_score = copy.copy(self.current_game_state.get_scores())
                id_player_error = self.current_game_state.get_next_player().get_id()
                temp_score.pop(id_player_error)
                self.winner = self.compute_winner(temp_score)
                self.current_game_state.get_scores()[id_player_error] = float(sys.maxsize)
                scores = self.get_scores()
                for key in scores.keys() :
                    logger.info(f"{key} - {scores[key]}")
                for player in self.get_winner() :
                    logger.info(f"Winner - {player.get_name()}")

                await self.emitter.sio.emit("done",json.dumps(self.get_scores()))

                return self.winner

            logger.info(f"Current game state: \n{self.current_game_state.get_rep()}")

            await self.emitter.sio.emit(
                "play",
                json.dumps(self.current_game_state.to_json(),default=lambda x:x.to_json()),
            )

        self.winner = self.compute_winner(self.current_game_state.get_scores())
        scores = self.get_scores()
        for key in scores.keys() :
                    logger.info(f"{key} - {scores[key]}")
        for player in self.get_winner() :
            logger.info(f"Winner - {player.get_name()}")

        await self.emitter.sio.emit("done",json.dumps(self.get_scores()))

        return self.winner

    def record_game(self, listeners:Optional[List[EventSlave]]=None) -> None:
        """
        Starts a game and broadcasts its successive states.
        """
        self.emitter.start(self.play_game, self.players+(listeners if listeners else []))

    def update_log(self) -> None:
        """
        Updates the log file.

        Raises:
            MethodNotImplementedError: If the method is not implemented.
        """
        raise MethodNotImplementedError()

    def get_name(self) -> str:
        """
        Returns:
            str: The name of the game.
        """
        return self.name

    def get_game_state(self) -> GameState:
        """
        Returns:
            GameState: The current game state.
        """
        return self.current_game_state

    def get_json_path(self) -> str:
        """
        Returns:
            str: The path of the log file.
        """
        return self.log_level

    def get_winner(self) -> list[Player]:
        """
        Returns:
            Player: The winner(s) of the game.
        """
        return self.winner

    def get_scores(self) -> dict[int, float]:
        """
        Returns:
            Dict: The scores of the current state.
        """
        return self.current_game_state.get_scores()

    @abstractmethod
    def compute_winner(self, scores: dict[int, float]) -> list[Player]:
        """
        Computes the winner(s) of the game based on the scores.

        Args:
            scores (Dict[int, float]): The score for each player.

        Raises:
            MethodNotImplementedError: If the method is not implemented.

        Returns:
            Iterable[Player]: The list of player(s) who won the game.
        """
        raise MethodNotImplementedError()



from abc import abstractmethod
from typing import Any

from seahorse.utils.custom_exceptions import MethodNotImplementedError
from seahorse.utils.serializer import Serializable


class Representation(Serializable):
    """
    A class representing a game representation.

    Attributes:
        env (dict): The dictionary of the current state.
    """

    def __init__(self, env: dict) -> None:
        """
        Initializes a new instance of the Representation class.

        Args:
            env (dict): The dictionary of the current state.
        """
        self.env = env

    def get_env(self) -> dict:
        """
        Gets the dictionary of the environment.

        Returns:
            dict: The dictionary of the environment.
        """
        return self.env

    def find(self, to_find: Any) -> Any:
        """
        Finds a key directly in the environment.

        Args:
            to_find (Any): The key to find.

        Returns:
            Any: The value of the cell.
        """
        if to_find not in self.env.keys():
            return -1
        else:
            return self.env[to_find]

    @abstractmethod
    def __hash__(self) -> int:
        raise MethodNotImplementedError()

    @abstractmethod
    def __eq__(self, __value: object) -> bool:
        raise MethodNotImplementedError()

    @abstractmethod
    def __str__(self) -> str:
        raise MethodNotImplementedError()



class AlreadyRunningError(Exception):
    """Thrown when trying to start somethin twice
    """

    def __init__(self,  message: str = "Trying to start something twice !"):
        self.message = message
        super().__init__(message)


class NotRunningError(Exception):
    """Thrown when trying to stop somethin twice
    """

    def __init__(self,  message: str = "Trying to stop something twice !"):
        self.message = message
        super().__init__(message)

class PlayerDuplicateError(Exception):
    """Thrown when trying to stop somethin twice
    """

    def __init__(self,  message: str = "Duplicates in the players list."):
        self.message = message
        super().__init__(message)


class SeahorseTimeoutError(Exception):
    """Thrown when trying to modify an expired element
    """

    def __init__(self,  message: str = "Trying to modify an expired element ! Time is out !"):
        self.message = message
        super().__init__(message)


class TimerNotInitializedError(Exception):
    """Thrown when trying to use timer utilities before timer initialization
    """

    def __init__(self,  message: str = "Timer not initialized."):
        self.message = message
        super().__init__(message)


class MethodNotImplementedError(Exception):
    """Thrown when trying to use a method not implemented
    """

    def __init__(self,  message: str = "Method not implemented."):
        self.message = message
        super().__init__(message)


class ActionNotPermittedError(Exception):
    """Thrown when trying to generate an action that's not permitted
    """

    def __init__(self,  message: str = "Action not permitted"):
        self.message = message
        super().__init__(message)

class StopAndStartError(Exception):
    """Thrown when trying to generate an action that's not permitted
    """

    def __init__(self,  message: str = "Timer was tampered with and exceeded the tolerancy set in GameMaster."):
        self.message = message
        super().__init__(message)

class ConnectionProblemError(Exception):
    """Thrown when trying to generate an action that's not permitted
    """

    def __init__(self,  message: str = "Connection problem"):
        self.message = message
        super().__init__(message)


class NoTournamentFailError(Exception):
    """Thrown when trying to generate an action that's not permitted
    """

    def __init__(self,  message: str = "Tournament problem : tournament is none, please connect to an existing tournament or create a tournament"):
        self.message = message
        super().__init__(message)



from __future__ import annotations

from typing import TYPE_CHECKING

from seahorse.utils.serializer import Serializable

if TYPE_CHECKING:
    from seahorse.game.game_state import GameState


class Action(Serializable):
    """
    A class representing an action in the game.

    Attributes:
        past_gs (GameState): The past game state.
        new_gs (GameState): The new game state.
    """

    def __init__(self, current_game_state: GameState, next_game_state: GameState) -> None:
        """
        Initializes a new instance of the Action class.

        Args:
            past_gs (GameState): The past game state.
            new_gs (GameState): The new game state.
        """
        self.current_game_state = current_game_state
        self.next_game_state = next_game_state

    def get_current_game_state(self) -> GameState:
        """
        Returns the past game state.

        Returns:
            GameState: The past game state.
        """
        return self.current_game_state

    def get_next_game_state(self) -> GameState:
        """
        Returns the new game state.

        Returns:
            GameState: The new game state.
        """
        return self.next_game_state

    def __hash__(self) -> int:
        return hash((hash(self.get_next_game_state()), hash(self.get_current_game_state())))

    def __eq__(self, value: object) -> bool:
        return hash(self) == hash(value)

    def __str__(self) -> str:
        return "From:\n" + self.get_current_game_state().get_rep().__str__() + "\nto:\n" + self.get_next_game_state().get_rep().__str__()

    def to_json(self) -> dict:
        return self.__dict__



from __future__ import annotations

import json
from typing import TYPE_CHECKING

from seahorse.game.representation import Representation
from seahorse.utils.serializer import Serializable

if TYPE_CHECKING:
    from seahorse.player.player import Player


class Piece(Serializable):
    """
    A class representing a piece in the game.

    Attributes:
        piece_type (str): The type of the piece.
        owner_id (int): The ID of the player who possesses the piece.
    """

    def __init__(self, piece_type: str, owner: Player | None = None, owner_id: int=-1) -> None:
        """
        Initializes a new instance of the Piece class.

        Args:
            piece_type (str): The type of the piece.
            owner (Player): The player who possesses the piece.
        """
        self.piece_type = piece_type
        if owner is None:
            self.owner_id = owner_id
        else:
            self.owner_id = owner.get_id()

    def get_type(self) -> str:
        """
        Gets the type of the piece.

        Returns:
            str: The type of the piece.
        """
        return self.piece_type

    def get_owner_id(self) -> int:
        """
        Gets the ID of the owner of the piece.

        Returns:
            int: The ID of the owner.
        """
        return self.owner_id

    def copy(self) -> Piece:
        """
        Creates a copy of the piece.

        Returns:
            Piece: A copy of the piece.
        """
        return Piece(self.piece_type, None)

    def __hash__(self) -> int:
        return hash((hash(self.get_type()), hash(self.owner_id)))

    def __eq__(self, __value: object) -> bool:
        return hash(self) == hash(__value)

    #def __str__(self) -> str:
    #    return self.get_type()

    def to_json(self) -> str:
        return self.__dict__

    @classmethod
    def from_json(cls,data) -> str:
        return cls(**json.loads(data))


class Board(Representation):
    """
    A class representing the game board.

    Attributes:
        env (dict[Tuple[int], Piece]): The environment dictionary composed of pieces.
        dimensions (list[int]): The dimensions of the board.
    """

    def __init__(self, env: dict[tuple[int], Piece], dim: list[int]) -> None:
        """
        Initializes a new instance of the Board class.

        Args:
            env (dict[Tuple[int], Piece]): The environment dictionary composed of pieces.
            dim (list[int]): The dimensions of the board.
        """
        super().__init__(env)
        self.dimensions = dim

    def get_dimensions(self) -> list[int]:
        """
        Gets the dimensions of the board.

        Returns:
            list[int]: The list of dimensions.
        """
        return self.dimensions

    def get_pieces_player(self, owner: Player) -> tuple[int, list[Piece]]:
        """
        Gets the pieces owned by a specific player.

        Args:
            owner (Player): The player specified.

        Returns:
            Tuple[int, list[Piece]]: The number of pieces owned by the player and the list of their pieces.
        """
        pieces_list = []
        number = 0
        for key in self.env.keys():
            if self.env[key].get_owner_id() == owner.get_id():
                number += 1
                pieces_list.append(key)
        return number, pieces_list

    def __hash__(self):
        return hash(frozenset([(hash(pos), hash(piece)) for pos, piece in self.env.items()]))

    def __eq__(self, __value: object) -> bool:
        return hash(self) == hash(__value)

    def __str__(self) -> str:
        dim = self.get_dimensions()
        to_print = ""
        for i in range(dim[0]):
            for j in range(dim[1]):
                if self.get_env().get((i, j), -1) != -1:
                    to_print += str(self.get_env().get((i, j)).get_type()) + " "
                else:
                    to_print += "_ "
            to_print += "\n"
        return to_print



"""Example Google style docstrings.

This module demonstrates documentation as specified by the `Google Python
Style Guide`_. Docstrings may extend over multiple lines. Sections are created
with a section header and a colon followed by a block of indented text.

Example:
    Examples can be given using either the ``Example`` or ``Examples``
    sections. Sections support any reStructuredText formatting, including
    literal blocks::

        python example_google.py

Section breaks are created by resuming unindented text. Section breaks
are also implicitly created anytime a new section starts.

Attributes:
    module_level_variable1 (int): Module level variables may be documented in
        either the ``Attributes`` section of the module docstring, or in an
        inline docstring immediately following the variable.

        Either form is acceptable, but the two should not be mixed. Choose
        one convention to document module level variables and be consistent
        with it.

Todo:
    * For module TODOs
    * You have to also use ``sphinx.ext.todo`` extension

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""



import json
from collections.abc import Coroutine
import time
from typing import Any, Optional

from loguru import logger
from seahorse.game.action import Action

from seahorse.game.game_state import GameState
from seahorse.game.io_stream import EventMaster, EventSlave, event_emitting, remote_action
from seahorse.player.player import Player
from seahorse.utils.custom_exceptions import MethodNotImplementedError
from seahorse.utils.gui_client import GUIClient
from seahorse.utils.serializer import Serializable


class RemotePlayerProxy(Serializable,EventSlave):
    """
    A class representing a remote player proxy.

    Attributes:
        mimics (type[Player]): The player type to mimic.
        sid: The session ID.
    """

    def __init__(self, mimics: type[Player], *args, **kwargs) -> None:
        """
        Initializes a new instance of the RemotePlayerProxy class.

        Args:
            mimics (type[Player]): The player type to mimic.
            *args: Additional arguments.
            **kwargs: Additional keyword arguments.
        """
        self.mimics = mimics(*args, **kwargs)
        self.activate(wrapped_id=self.mimics.get_id())
        self.id = self.mimics.id
        self.sid = None

    @remote_action("turn")
    def play(self, *,current_state: GameState) -> Action:
        """
        Plays a move.

        Args:
            current_state: The game state.

        Returns:
            Action: The action resulting from the move.
        """
        pass

    async def listen(self,**_) -> Coroutine[Any, Any, None]:
        """
        Fires up the listening process

        Returns:
            Coroutine: A coroutine object.
        """
        idmap = await EventMaster.get_instance().wait_for_identified_client(self.name,self.id)
        self.sid = idmap["sid"]

    def __getattr__(self, attr):
        return getattr(self.mimics, attr)

    def __hash__(self) -> int:
        return hash(self.sid)

    def __eq__(self, __value: object) -> bool:
        return hash(self) == hash(__value)

    def to_json(self) -> str:
        return str(self.wrapped_id)


class LocalPlayerProxy(Serializable,EventSlave):
    """
    A class representing a local player proxy.

    Attributes:
        wrapped_player (Player): The wrapped player object.

    Methods:
        play(current_state: GameState) -> Action: Plays a move.
    """

    def __init__(self, wrapped_player: Player,gs:type=GameState) -> None:
        """
        Initializes a new instance of the LocalPlayerProxy class.

        Args:
            wrapped_player (Player): The player object to wrap.
        """
        self.wrapped_player = wrapped_player
        self.activate(self.wrapped_player.name,wrapped_id=wrapped_player.get_id())
        @self.sio.on("turn")
        async def handle_turn(*data):
            logger.info(f"{self.wrapped_player.name} is playing")
            logger.debug(f"Data received : {data}")
            logger.debug(f"Deserialized data : \n{gs.from_json(data[0],next_player=self)}")
            action = await self.play(gs.from_json(data[0],next_player=self))
            logger.info(f"{self.wrapped_player} played the following action : \n{action}")

        @self.sio.on("update_id")
        async def update_id(data):
            logger.debug("update_id received",json.loads(data)["new_id"])
            self.wrapped_player.id = json.loads(data)["new_id"]

    @event_emitting("action")
    def play(self, current_state: GameState) -> Action:
        """
        Plays a move.

        Args:
            current_state (GameState): The current game state.

        Returns:
            Action: The action resulting from the move.
        """
        return self.compute_action(current_state=current_state)

    def __getattr__(self, attr):
        return getattr(self.wrapped_player, attr)

    def __hash__(self) -> int:
        return hash(self.wrapped_player)

    def __eq__(self, __value: object) -> bool:
        return hash(self) == hash(__value)

    def __str__(self) -> str:
        return f"Player {self.wrapped_player.get_name()} has ID {self.wrapped_player.get_id()}."

    def to_json(self) -> dict:
        return self.wrapped_player.to_json()

class InteractivePlayerProxy(LocalPlayerProxy):
    """Proxy for interactive players, 
       inherits from `LocalPlayerProxy`
    """
    def __init__(self, mimics: Player, gui_path:Optional[str]=None, *args, **kwargs) -> None:
        """

        Args:
            mimics (type[Player]): A wrapped player, the internal logic will be overridden by an interactive one
            gui_path (str, optional): If the interaction is supposed to happen on the host machine, provide a GUI path to start it up. Defaults to None.
        """
        super().__init__(mimics, *args, **kwargs)
        self.wrapped_player.player_type = "interactive"
        self.path = gui_path
        self.shared_sid = None
        self.sid = None

    async def play(self, current_state: GameState) -> Action:
        if self.shared_sid and not self.sid:
            self.sid=self.shared_sid.sid
        while True:
            data = json.loads(await EventMaster.get_instance().wait_for_event(self.sid,"interact",flush_until=time.time()))
            try:
                action = current_state.convert_light_action_to_action(data)
            except MethodNotImplementedError:
                #TODO: handle this case
                action = Action.from_json(data)

            if action in current_state.get_possible_actions():
                break
            else:
                await EventMaster.get_instance().sio.emit("ActionNotPermitted",None)
        return action

    async def listen(self, master_address, *, keep_alive: bool) -> None:
        if not self.shared_sid:
            await super().listen(master_address, keep_alive=keep_alive)
            embedded_client = GUIClient(path=self.path)
            await embedded_client.listen()
            self.sid = embedded_client.sid

    def share_sid(self,proxy:"InteractivePlayerProxy"):
        self.shared_sid=proxy




import builtins
import functools
import time
from typing import Any

from loguru import logger

from seahorse.utils.custom_exceptions import (
    AlreadyRunningError,
    NotRunningError,
    SeahorseTimeoutError,
    TimerNotInitializedError,
)


class TimeMaster:
    __instance = None

    class Timer:
        def __init__(self,time_limit:float=1e9):
            self._time_limit = time_limit
            self._remaining_time = time_limit
            self._last_timestamp = None
            self._is_running = False

        def start_timer(self) -> float:
            """Starts the timer

            Raises:
                AlreadyRunningException: when trying to start twice.
            """
            if self._is_running:
                raise AlreadyRunningError()

            self._last_timestamp = time.time()

            self._is_running = True

            return self._last_timestamp

        def is_running(self) -> bool:
            """
            Is the timer running ?

            Returns:
                bool: `True` if the timer is running, `False` otherwise
            """
            return self._is_running

        def get_time_limit(self):
            """
            Get the limit set in `set_time_limit()`
            """
            return self._time_limit

        def get_last_timestamp(self):
            """
            Get the last timestamp set at start_timer()
            """
            return self._last_timestamp

        def get_remaining_time(self) -> float:
            """Gets the timer's remaining time

            Returns:
                float: the remaining time
            """
            if self._is_running:
                return self._remaining_time - (time.time() - self._last_timestamp)
            else:
                return self._remaining_time

        def stop_timer(self) -> float:
            """Pauses the timer

            Raises:
                NotRunningException: when the timer isn't running

            Returns:
                float: remaining time
            """
            if not self._is_running:
                raise NotRunningError()

            self._remaining_time = self._remaining_time - (time.time() - self._last_timestamp)

            self._is_running = False
            return self._remaining_time


        def is_locked(self) -> bool:
            """Is the time credit expired ?

            Returns:
                bool: `True` if expired `False` otherwise
            """
            logger.info(f"time : {self.get_remaining_time()}")
            return self.get_remaining_time() <= 0

    @staticmethod
    def get_instance()->"TimeMaster":
        if TimeMaster.__instance is None:
            TimeMaster.__instance=TimeMaster()
        return TimeMaster.__instance

    @classmethod
    def register_timer(cls: "TimeMaster", linked_instance: Any, time_limit:float=1e9):
        pid = linked_instance.__dict__.get("id",builtins.id(linked_instance))
        cls.get_instance().__time_register[pid]=cls.get_instance().__time_register.get(pid,TimeMaster.Timer(time_limit))

    @classmethod
    def get_timer(cls: "TimeMaster", linked_instance: Any)-> Timer:
        return cls.get_instance().__time_register.get(linked_instance.__dict__.get("id",builtins.id(linked_instance)))


    def __init__(self):
        if TimeMaster.__instance is not None:
            msg = "Trying to initialize multiple instances of TimeMaster, this is forbidden to avoid side-effects.\n Call TimeMaster.get_instance() instead."
            raise NotImplementedError(msg)
        else:
            self.__time_register={}

class TimeMixin:
    """
    When implemented allows any object to keep track of time

    Example usage:
    ```
        import time
        class MyTimedObject(TimeMixin):
            def __init__(self):
                self.myattr = 2

            x = MyTimedObject()
            x.set_time_limit(10)
            x.start_timer()
            time.sleep(11)
            x.myattr=5  # raises SeahorseTimeoutException

    ```
    """

    def init_timer(self, time_limit: int) -> None:
        """
        Initializes the time credit of the instance

        Doesn't start the timer yet ! Call `start_timer()`.

        Args:
            time_limit (int): max time before locking all methods of the class
        """
        TimeMaster.register_timer(self,time_limit)

    def start_timer(self) -> float:
        """Starts the timer

        Raises:
            AlreadyRunningException: when trying to start twice.
        """
        if TimeMaster.get_timer(self) is None:
            raise TimerNotInitializedError
        return TimeMaster.get_timer(self).start_timer()

    def is_running(self) -> bool:
        """
        Is the timer running ?

        Returns:
            bool: `True` if the timer is running, `False` otherwise
        """
        if TimeMaster.get_timer(self) is None:
            raise TimerNotInitializedError
        return TimeMaster.get_timer(self).is_running()

    def get_time_limit(self):
        """
        Get the limit set in `set_time_limit()`
        """
        if TimeMaster.get_timer(self) is None:
            raise TimerNotInitializedError
        return TimeMaster.get_timer(self).get_time_limit()

    def get_remaining_time(self) -> float:
        """Gets the timer's remaining time

        Returns:
            float: the remaining time
        """
        if TimeMaster.get_timer(self) is None:
            raise TimerNotInitializedError
        return TimeMaster.get_timer(self).get_remaining_time()

    def get_last_timestamp(self) -> float:
        """Gets the timer's last recorded timestamp at which it was started

        Returns:
            float: the timestamp
        """
        if TimeMaster.get_timer(self) is None:
            raise TimerNotInitializedError
        return TimeMaster.get_timer(self).get_last_timestamp()

    def stop_timer(self) -> float:
        """Pauses the timer

        Raises:
            NotRunningException: when the timer isn't running

        Returns:
            float: remaining time
        """
        if TimeMaster.get_timer(self) is None:
            raise TimerNotInitializedError
        return TimeMaster.get_timer(self).stop_timer()


    def is_locked(self) -> bool:
        """Is the time credit expired ?

        Returns:
            bool: `True` if expired `False` otherwise
        """
        if TimeMaster.get_timer(self) is None:
            raise TimerNotInitializedError
        return TimeMaster.get_timer(self).is_locked()

    def __setattr__(self, __name: str, value: Any) -> None:
        """_summary_

        Args:
            Inherited from object
        Raises:
            TimeoutException: prevents modification after timout

        """
        try:
            if TimeMaster.get_timer(self) and self.is_locked():
                raise SeahorseTimeoutError()
            else:
                self.__dict__[__name] = value
        except Exception as e:
            raise e

def timed_function(fun):
    """
    Decorator to prevent using a function after object's timeout.
    Args:
        fun (_type_): wrapped function

    Raises:
        TimerNotInitializedError: _description_
        Exception: _description_
        SeahorseTimeoutError: _description_

    Returns:
        Callable[...]: wrapper
    """
    @functools.wraps(fun)
    def wrapper(self, *args, **kwargs):
        r = fun(self, *args, **kwargs)
        if TimeMaster.get_timer(self) is None:
            raise TimerNotInitializedError
        elif(self.is_locked()):
            raise SeahorseTimeoutError()
        return r
    return wrapper



from typing import Any, Dict, List
import unittest
import copy
from random import sample

from seahorse.game.game_layout.board import Piece, Board
from seahorse.game.representation import Representation
from seahorse.player.player import Player
from seahorse.game.game_state import GameState
from seahorse.game.action import Action


class Dummy_GameState(GameState):

    def __init__(self, scores: Dict[int, Any], next_player: Player, players: List[Player], rep: Representation) -> None:
        super().__init__(scores, next_player, players, rep)

    def generate_possible_actions(self):
        list_rep = []
        current_rep = self.get_rep()
        next_player = self.get_next_player()
        for i in range(current_rep.get_dimensions()[0]):
            for j in range(current_rep.get_dimensions()[1]):
                if not current_rep.get_env().get((i, j)):
                    copy_rep = copy.deepcopy(current_rep)
                    copy_rep.get_env()[(i, j)] = Piece(piece_type="Added", owner=next_player)
                    list_rep.append(copy.deepcopy(copy_rep))
        poss_actions = {
            Action(
                self,
                Dummy_GameState(
                    self.get_scores(),
                    self.compute_next_player(),
                    self.players,
                    valid_next_rep,
                ),
            )
            for valid_next_rep in list_rep
        }

        return poss_actions


class test_case(unittest.TestCase):

    def setUp(self):
        self.board = Board(env={}, dim=[3, 3])

        self.player1 = Player("Thomas")
        self.player2 = Player(id=42)

        self.piece1 = Piece("A")
        self.piece2 = Piece("B", self.player2)
        self.piece3 = Piece("C", self.player1)

        self.current_gs = Dummy_GameState(scores={self.player1.get_id():1, self.player2.get_id():0}, next_player=self.player1, players=[self.player1, self.player2], rep=self.board)

    def test_id(self):
        assert self.player1.get_id() == id(self.player1)
        assert self.player2.get_id() == 42
        assert self.piece1.get_owner_id() == -1
        assert self.piece2.get_owner_id() == self.player2.get_id()

    def test_board(self):
        assert self.board.get_dimensions() == [3,3]
        assert self.board.get_env() == {}
        assert self.board.get_pieces_player(self.player1) == (0, [])
        self.board.env[(0, 1)] = self.piece1
        self.board.env[(2, 1)] = self.piece2
        self.board.env[(2, 2)] = self.piece3
        assert self.board.get_pieces_player(self.player1) == (1, [(2, 2)])
        assert self.board.get_pieces_player(self.player2) == (1, [(2, 1)])

    def test_gamestate(self):
        assert self.current_gs.compute_next_player() == self.player2
        assert self.current_gs.get_player_score(self.player1) == 1
        self.board.env[(0, 1)] = self.piece1
        possible_actions = self.current_gs.generate_possible_actions()
        assert len(possible_actions) == self.current_gs.get_rep().get_dimensions()[0]*self.current_gs.get_rep().get_dimensions()[1] - 1
        self.board.env[(2, 1)] = self.piece2
        self.board.env[(2, 2)] = self.piece3
        possible_actions = self.current_gs.generate_possible_actions()
        assert len(possible_actions) == self.current_gs.get_rep().get_dimensions()[0]*self.current_gs.get_rep().get_dimensions()[1] - 3



from __future__ import annotations

import builtins
from abc import abstractmethod
from typing import TYPE_CHECKING

from seahorse.game.action import Action
from seahorse.game.time_manager import TimeMixin, timed_function
from seahorse.utils.custom_exceptions import MethodNotImplementedError
from seahorse.utils.serializer import Serializable

if TYPE_CHECKING:
    from seahorse.game.game_state import GameState


class Player(Serializable,TimeMixin):
    """
    A base class representing a player in the game.

    Attributes:
        id (int): The ID of the player.
        name (str) : the name of the player
    """

    def __init__(self, name: str = "bob", time_limit: float = 1e6,*,id:int | None = None,**_) -> None:
        """
        Initializes a new instance of the Player class.

        Args:
            name (str, optional): The name of the player. Defaults to "bob".
            time_limit (float, optional): The time limit for the player's moves. Defaults to 1e6.
            hard_id (int, optional, keyword-only): Set the player's id in case of distant loading
        """
        self.name = name
        if id is None:
            self.id = builtins.id(self)
        else:
            self.id = id
        self.init_timer(time_limit)


    @timed_function
    def play(self, current_state: GameState) -> Action:
        """
        Implements the player's logic and calls compute_action with minimal information.

        Args:
            current_state (GameState): The current game state.

        Raises:
            MethodNotImplementedError: If the method is not implemented in the derived class.

        Returns:
            Action: The resulting action.
        """
        # TODO: check score ????
        return self.compute_action(current_state=current_state)

    @abstractmethod
    def compute_action(self, **kwargs) -> Action:
        """
        Should be dedicated to adversarial search.

        Args:
            **kwargs: Additional arguments.

        Raises:
            MethodNotImplementedError: If the method is not implemented in the derived class.

        Returns:
            Action: The action to play.
        """
        raise MethodNotImplementedError()

    def get_id(self) -> int:
        """
        Returns:
            int: The ID of the player.
        """
        return self.id

    def get_name(self) -> str:
        """
        Returns:
            str: The name of the player.
        """
        return self.name

    def __str__(self) -> str:
        """
        Returns a string representation of the Player object.

        Returns:
            str: The string representation.
        """
        return f"Player {self.get_name()}({self.get_id()})"



from __future__ import annotations

import asyncio
import csv
import math
from sys import platform

import challonge
from split import chop

from seahorse.utils.custom_exceptions import ConnectionProblemError, NoTournamentFailError


class ChallongeTournament:
    """
    A class to interact with the Challonge tournament platform.

    Attributes:
        id_challonge (str): The Challonge ID.
        keypass_challonge (str): The Challonge API key.
        game_name (str): The name of the game.
        log_level (str): The log file.
    """

    def __init__(self, id_challonge: str, keypass_challonge: str, game_name: str, log_level: str|None, log_file: str = "log.txt") -> None:
        """
        Initializes a new instance of the ChallongeTournament class.

        Args:
            id_challonge (str): The Challonge ID.
            keypass_challonge (str): The Challonge API key.
            game_name (str): The name of the game.
            log_level (str): The log file. Default is None.
        """
        self.id_challonge = id_challonge
        self.keypass_challonge = keypass_challonge
        self.game_name = game_name
        self.log_level = log_level
        self.user = None
        self.tournament = None
        self.log_file = log_file
        self.created = False

    async def create_tournament(self, tournament_name: str, tournament_url: str, csv_file: str, sep: str = ",") -> None:
        """
        Creates a new tournament on Challonge and adds participants from a CSV file.

        Args:
            tournament_name (str): The name of the tournament.
            tournament_url (str): The URL of the tournament.
            csv_file (str): The path to the CSV file containing participant names.
            sep (str): The delimiter used in the CSV file. Default is ",".

        Returns:
            None
        """
        self.user = await challonge.get_user(self.id_challonge, self.keypass_challonge)
        self.tournament = await self.user.create_tournament(name=tournament_name, url=tournament_url)
        with open(csv_file) as csvfile :
            spamreader = csv.reader(csvfile, delimiter=sep)
            for line in spamreader :
                for name in line :
                    await self.tournament.add_participant(str(name))
        self.created = True

    async def connect_tournament(self, tournament_name: str) -> None:
        """
        Connects to an existing tournament on Challonge.

        Args:
            tournament_name (str): The name of the tournament.

        Returns:
            None

        Raises:
            ConnectionProblemError: If the connection to the tournament fails.
        """
        self.user = await challonge.get_user(self.id_challonge, self.keypass_challonge)
        my_tournaments = await self.user.get_tournaments()
        for t in my_tournaments:
            if t.name == tournament_name :
                self.tournament = t
                return
        raise ConnectionProblemError()

    def retrieve_scores(self, match) -> str:
        """
        Retrieves the scores from a match.

        Args:
            match: The match object.

        Returns:
            str: The scores as a string.
        """
        if not match.scores_csv :
            return match.scores_csv
        return match.scores_csv + ","

    def retrieve_winners(self, scores: str, p1, p2, minormax: str) -> list :
        """
        Retrieves the winners from the scores.

        Args:
            scores (str): The scores as a string.
            p1: The participant object of player 1.
            p2: The participant object of player 2.

        Returns:
            list: A list of winners.
        """
        result = []
        if not scores :
            return result
        list_scores = scores[:-1].split(",")
        for score in list_scores:
            s1, s2 = score.split("-")
            if minormax == "max" :
                if s1 > s2 :
                    result.append(p1)
                elif s1 < s2 :
                    result.append(p2)
            elif minormax == "min" :
                if s1 > s2 :
                    result.append(p2)
                elif s1 < s2 :
                    result.append(p1)
        return result

    def invert_score(self, score: str) -> str:
        """
        Inverts the score.

        Args:
            score (str): The score as a string.

        Returns:
            str: The inverted score.
        """
        list_score = score[:-1].split("-")
        return list_score[1] + "-" + list_score[0] + ","

    def get_participant_winner(self, winner: str, p1, p2):
        """
        Gets the participant object of the winner.

        Args:
            winner (str): The name of the winner.
            p1: The participant object of player 1.
            p2: The participant object of player 2.

        Returns:
            The participant object of the winner.
        """
        if winner == p1.name :
            return p1
        else :
            return p2

    async def play_round(self,name1: str, name2: str, port: int, folder_player: str) -> tuple[str, str]:
        """
        Plays a round of the tournament.

        Args:
            name1 (str): The name of player 1.
            name2 (str): The name of player 2.
            port (int): The port number.
            folder_player (str): The folder containing the player scripts.

        Returns:
            tuple[str, str]: A tuple containing the score and the winner.
        """
        if platform == "win32" :
            cmd = "python " + self.game_name + ".py" + " -t local -p " + str(port) + " " + folder_player+name1 + " " + folder_player+name2
        else :
            cmd = "python3 " + self.game_name + ".py" + " -t local -p " + str(port) + " " + name1 + " " + name2
        process = await asyncio.create_subprocess_shell(cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        stdout, stderr = await process.communicate()
        score1 = stderr.decode("utf-8").split("\n")[-4].split(" - ")[-1]
        score2 = stderr.decode("utf-8").split("\n")[-3].split(" - ")[-1]
        winner = stderr.decode("utf-8").split("\n")[-2].split(" - ")[-1]
        score = str(math.floor(abs(float(score1)))) + "-" + str(math.floor(abs(float(score2)))) + ","
        winner = str(winner)
        with open(self.log_file,"a+") as f :
            f.write(stderr.decode(encoding="utf-8"))
        f.close()
        return score, winner

    async def play_match(self, match, port: int, rounds: int, folder_player: str, minormax: str) -> None:
        """
        Plays a match of the tournament.

        Args:
            match: The match object.
            port (int): The port number.
            rounds (int): The number of rounds.
            folder_player (str): The folder containing the player scripts.

        Returns:
            None
        """
        if match.completed_at is None :
            p1 = await self.tournament.get_participant(match.player1_id)
            p2 = await self.tournament.get_participant(match.player2_id)
            already_played = 0
            if match.underway_at is None :
                await match.mark_as_underway()
                scores = ""
                winners = []
            else :
                scores = self.retrieve_scores(match)
                winners = self.retrieve_winners(scores, p1, p2, minormax)
                already_played = len(winners)
            for r in range(already_played, rounds) :
                if r % 2 == 0 :
                    score, winner = await self.play_round(p1.name, p2.name, port, folder_player)
                    scores += score
                    winners.append(self.get_participant_winner(winner, p1, p2))
                else :
                    score, winner = await self.play_round(p2.name, p1.name, port, folder_player)
                    scores += self.invert_score(score)
                    winners.append(self.get_participant_winner(winner, p1, p2))
                await match.report_live_scores(scores[:-1])
            if match.group_id is not None :
                await match._report(scores[:-1], max(winners,key=winners.count).group_player_ids[0])
            else :
                await match._report(scores[:-1], max(winners,key=winners.count).id)
            await match.unmark_as_underway()

    async def run(self, folder_player: str, rounds: int = 1, nb_process: int = 2, minormax: str = "max") -> None:
        """
        Runs the tournament.

        Args:
            folder_player (str): The folder containing the player scripts.
            rounds (int): The number of rounds. Default is 1.
            nb_process (int): The number of parallel processes. Default is 2.

        Returns:
            None

        Raises:
            NoTournamentFailError: If there is no tournament.
        """
        if self.tournament is not None :
            if self.created :
                await self.tournament.start()
            matches = await self.tournament.get_matches()
            dict_round = {}
            for match in matches :
                if match.group_id is None :
                    if dict_round.get(match.round,False) :
                        dict_round[match.round] += [match]
                    else :
                        dict_round[match.round] = [match]
                elif dict_round.get(match.group_id,False) :
                    dict_round[match.group_id] += [match]
                else :
                    dict_round[match.group_id] = [match]
            for key in dict_round.keys() :
                port = 16000
                for matches in list(chop(nb_process, dict_round[key])) :
                    list_jobs_routines = [asyncio.create_task(self.play_match(match, port+i, rounds, folder_player, minormax)) for i, match in enumerate(matches)]
                    await asyncio.gather(*list_jobs_routines)
            if self.created :
                await self.tournament.finalize()
        else :
            raise NoTournamentFailError()



from abc import abstractmethod
from itertools import cycle
from typing import Any

from seahorse.game.action import Action
from seahorse.game.representation import Representation
from seahorse.player.player import Player
from seahorse.utils.custom_exceptions import MethodNotImplementedError
from seahorse.utils.serializer import Serializable


class GameState(Serializable):
    """
    A class representing the game state.

    Attributes:
        scores (Dict[int, Any]): The scores of the state for each player.
        next_player (Player): The next player to play.
        players (List[Player]): The list of players.
        rep (Representation): The representation of the game.
    """

    def __init__(self, scores: dict[int, Any], next_player: Player, players: list[Player], rep: Representation) -> None:
        """
        Initializes a new instance of the GameState class.

        Args:
            scores (Dict[int, Any]): The scores of the state for each player.
            next_player (Player): The next player to play.
            players (List[Player]): The list of players.
            rep (Representation): The representation of the game.
        """
        self.scores = scores
        self.next_player = next_player
        self.players = players
        self.rep = rep
        self._possible_actions = None

    def get_player_score(self, player: Player) -> float:
        """
        Gets a player's score.

        Args:
            player (Player): The player.

        Returns:
            float: The score.
        """
        return self.scores[player.get_id()]

    def get_next_player(self) -> Player:
        """
        Returns the next player.

        Returns:
            Player: The next player.
        """
        return self.next_player

    def compute_next_player(self) -> Player:
        """
        Computes the next player.

        Returns:
            Player: The next player.
        """
        current = self.next_player
        curr_id = self.players.index(current)
        return next(cycle(self.players[curr_id + 1 :] + self.players[:curr_id]))

    def get_scores(self) -> dict[int, float]:
        """
        Returns the scores.

        Returns:
            Dict: The player ID to score mapping.
        """
        return self.scores

    def get_players(self) -> list[Player]:
        """
        Returns the list of players.

        Returns:
            List[Player]: The list of players.
        """
        return self.players

    def get_rep(self) -> Representation:
        """
        Returns the representation of the game.

        Returns:
            Representation: The game representation.
        """
        return self.rep

    def get_possible_actions(self) -> frozenset[Action]:
        """
        Returns a copy of the possible actions from this state.
        The first call triggers the `generate_possible_actions` method.

        Returns:
            FrozenSet[Action]: The possible actions.
        """
        # Lazy loading
        if self.is_done():
            return frozenset()
        if self._possible_actions is None:
            self._possible_actions = frozenset(self.generate_possible_actions())
        return self._possible_actions

    def check_action(self, action: Action) -> bool:
        """
        Checks if an action is feasible.

        Args:
            action (Action): The action to check.

        Returns:
            bool: True if the action is feasible, False otherwise.
        """
        if action in self.get_possible_actions():
            return True
        return False

    @abstractmethod
    def convert_light_action_to_action(self,data) ->  Action :
        raise MethodNotImplementedError()

    @abstractmethod
    def generate_possible_actions(self) -> set[Action]:
        """
        Generates a set of all possible actions from this game state.

        Returns:
            Set[Action]: A set of possible actions.

        Raises:
            MethodNotImplementedError: If the method is not implemented.
        """
        raise MethodNotImplementedError()

    @abstractmethod
    def compute_scores(self, next_rep: Representation) -> dict[int, Any]:
        """
        Computes the scores of each player.

        Args:
            next_rep (Representation): The next representation.

        Returns:
            Dict[int, Any]: The scores of each player.

        Raises:
            MethodNotImplementedError: If the method is not implemented.
        """
        raise MethodNotImplementedError()

    @abstractmethod
    def is_done(self) -> bool:
        """
        Indicates if the current GameState is final.

        Returns:
            bool: True if the state is final, False otherwise.

        Raises:
            MethodNotImplementedError: If the method is not implemented.
        """
        raise MethodNotImplementedError()

    def __hash__(self) -> int:
        return hash((hash(frozenset(self.scores.items())), hash(self.rep)))

    def __eq__(self, value: object) -> bool:
        return hash(self) == hash(value)

    def __str__(self) -> str:
        to_print = f"Current scores are {self.get_scores()}.\n"
        to_print += f"Next person to play is player {self.get_next_player().get_id()} ({self.get_next_player().get_name()}).\n"
        return to_print



from __future__ import annotations

import asyncio
import functools
import json
from collections import deque
import time
import re
from typing import TYPE_CHECKING, Any, Callable, Coroutine

import socketio
from aiohttp import web
from loguru import logger

from seahorse.game.action import Action
from seahorse.utils.serializer import Serializable

if TYPE_CHECKING:
    from seahorse.game.game_state import GameState


class EventSlave:

    def activate(self,
                 identifier:str | None = None,
                 wrapped_id:int | None = None,
                 *,
                 disconnected_cb: Callable[[None],None] | None = None
                 ) -> None:
        """
        Sets the listener up, binds handlers

        Args:
            identifier (str | None, optional): Must be a unique identifier. Defaults to None.
            wrapped_id (int | None, optional): If the eventSlave is bound to an instance a python native id might be associated. Defaults to None.
        """
        self.sio = socketio.AsyncClient()
        self.connected = False
        self.identifier = identifier
        self.wrapped_id = wrapped_id
        self.disconnected_cb = disconnected_cb

        @self.sio.event()
        async def connect():
            self.connected = True
            if self.identifier is not None:
                await self.sio.emit("identify", json.dumps(self.__dict__,default=lambda _:"_"))

        @self.sio.event
        def disconnect():
            self.connected = False


    async def listen(self,master_address:str,*,keep_alive:bool) -> None:
        """Fires up the listening process

        Args:
            master_address (str): the address to listen to
            keep_alive (bool): in standalone mode, this should be `True` to keep the asyncio process alive.
        """
        if not self.connected:
            await self.sio.connect(master_address)
        if keep_alive:
            while self.connected:
                await asyncio.sleep(.1)

def event_emitting(label:str):
    """Decorator to also send the function's output trough listening socket connexions

    Args:
        label (str): the type of event to emit
    """
    def meta_wrapper(fun: Callable[[Any],Action]):
        @functools.wraps(fun)
        async def wrapper(self:EventSlave,*args,**kwargs):
            out = fun(self,*args, **kwargs)
            await self.sio.emit(label,json.dumps(out.to_json(),default=lambda x:x.to_json()))
            return out

        return wrapper

    return meta_wrapper


def remote_action(label: str):
    """Proxy decorator to override an expected local behavior with a distant one
       *The logic in decorated function is ignored*
    Args:
        label (str): the time of event to emit to trigger the distant logic
    """
    def meta_wrapper(fun: Callable):
        @functools.wraps(fun)
        async def wrapper(self:EventSlave,current_state:GameState,*_,**__):
            await EventMaster.get_instance().sio.emit(label,json.dumps(current_state.to_json(),default=lambda x:x.to_json()),to=self.sid)
            out = await EventMaster.get_instance().wait_for_next_play(self.sid,current_state.players)
            return out

        return wrapper

    return meta_wrapper


class EventMaster:
    """
    Singleton for emitting events

    Attributes:


    Raises:
        NotImplementedError: when trying to initialize more than once

    """

    __instance = None

    @staticmethod
    def get_instance(game_state:type=Serializable,port: int = 8080,hostname: str="localhost") -> EventMaster:
        """Gets the instance object

        Args:
            n_clients (int, optional): the number of clients the instance is supposed to be listening, *ignored* if already initialized. Defaults to 1.
            game_state : class of a game state
            port (int, optional): the port to use. Defaults to 8080.

        Returns:
            object: _description_
        """
        if EventMaster.__instance is None:
            EventMaster(game_state=game_state, port=port, hostname=hostname)
        return EventMaster.__instance

    def __init__(self,game_state,port,hostname):
        if EventMaster.__instance is not None:
            msg = "Trying to initialize multiple instances of EventMaster, this is forbidden to avoid side-effects.\n Call EventMaster.get_instance() instead."
            raise NotImplementedError(msg)
        else:
            # Initializing attributes
            self.expected_clients = 0
            self.__n_clients_connected = 0
            self.__identified_clients = {}
            self.__open_sessions = set()
            self.__ident2sid = {}
            self.__sid2ident = {}
            self.__events = {}
            self.__game_state = game_state
            self.port = port
            self.hostname = hostname

            # Standard python-socketio server
            self.sio = socketio.AsyncServer(async_mode="aiohttp", async_handlers=True, cors_allowed_origins="*", ping_timeout=1e6)
            self.app = web.Application()

            # Starting asyncio stuff
            self.event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.event_loop)

            # Attaching the app
            self.sio.attach(self.app)

            self.runner = web.AppRunner(self.app)

            # Shutdown callback
            async def on_shutdown(_):
                for x in list(self.__open_sessions):
                    if x in self.__open_sessions:
                        await self.sio.disconnect(x)

            self.app.on_shutdown.append(on_shutdown)

            @self.sio.event()
            def connect(sid, *_):
                """
                    Handling incoming connections
                """
                self.__open_sessions.add(sid)
                self.__n_clients_connected += 1
                logger.info(f"Waiting for listeners {self.__n_clients_connected} out of {self.expected_clients} are connected.")

            @self.sio.event
            def disconnect(sid):
                logger.warning(f"Lost connection: {sid}")
                self.__n_clients_connected -= 1
                self.__open_sessions.remove(sid)
                if sid in self.__sid2ident.keys() and self.__sid2ident[sid] in self.__identified_clients:
                    logger.warning(f"Client identified as {self.__sid2ident[sid]} was lost.")
                    del self.__identified_clients[self.__sid2ident[sid]]

            @self.sio.on("*")
            async def catch_all(event,sid,data):
                self.__events[sid] = self.__events.get(sid,{})
                self.__events[sid][event] = self.__events[sid].get(event,deque())
                self.__events[sid][event].appendleft((time.time(),data))

            @self.sio.on("action")
            async def handle_play(sid,data):
                # TODO : cope with race condition "action" before "identify"
                try:
                    self.__identified_clients[self.__sid2ident[sid]]["incoming"].appendleft(data)
                # Plainly throw away packets that belong to disconnected clients
                except KeyError:
                    pass

            @self.sio.on("identify")
            async def handle_identify(sid,data):
                logger.info("Identifying a listener")
                logger.info(json.loads(data).get("identifier",0))
                logger.debug(f"Deserialized data {json.loads(data)}")
                data = json.loads(data)

                # TODO check presence of "id" in data
                idf = data.get("identifier",0)
                reg = r"^"+idf+r"(_duplicate_[0-9]+$|$)"
                if len(list(filter(lambda x:re.search(reg,x),self.__ident2sid.keys()))):
                    logger.warning("Two clients are using the same identifier, one of those will be ignored.")
                    idf = idf+"_duplicate_"+str(time.time())

                self.__ident2sid[idf]=sid
                self.__sid2ident[sid]=idf
                self.__identified_clients[idf]={"sid":sid,"id":data.get("wrapped_id",None),"incoming":deque(),"attached":False}


            # Setting the singleton instance
            EventMaster.__instance = self

    async def wait_for_next_play(self,sid:int,players:list) -> Action:
        """Waiting for the next play action, this function is blocking

        Args:
            sid (int): sid corresponding to the player to wait for
            players (list): the list of players

        Returns:
            Action: returns the received action
        """
        # TODO revise sanity checks to avoid critical errors
        logger.info(f"Waiting for next play from {self.__sid2ident[sid]}")
        while not len(self.__identified_clients[self.__sid2ident[sid]]["incoming"]):
            await asyncio.sleep(.1)
        logger.info("Action received")
        action = json.loads(self.__identified_clients[self.__sid2ident[sid]]["incoming"].pop())
        next_player_id = int(action["next_game_state"]["next_player"]["id"])
        next_player = next(iter(list(filter(lambda p:p.id==next_player_id,players))))

        past_gs = self.__game_state.from_json(json.dumps(action["current_game_state"]))
        past_gs.players = players
        new_gs = self.__game_state.from_json(json.dumps(action["next_game_state"]),next_player=next_player)
        new_gs.players = players


        return Action(past_gs,new_gs)

    async def wait_for_event(self,sid:int,label:str,*,flush_until:float | None=None) -> Coroutine:
        """Waits for an aribtrary event emitted by the connection identified by `sid`
           and labeled with `label`.
           One might want to ignore all events before a particular timestamp given in `flush_until`

        Args:
            sid (int): a socketio connexion identifier
            label (str): the event to wait for
            flush_until (float, optional): The timestamp treshold. Defaults to None.

        Returns:
            Coroutine: a promise yielding the data associated to the event 
        """
        while not len(self.__events.get(sid,{}).get(label,[])):
            await asyncio.sleep(.1)
        ts,data = self.__events[sid][label].pop()
        return data if (not flush_until) or ts>=flush_until else await self.wait_for_event(sid,label,flush_until=flush_until)

    async def wait_for_identified_client(self,name:str,local_id:int) -> str:
        """ Waits for an identified client (a player typically)

        Args:
            name (str): the name of the remote client
        Returns:
            str: the client sid
        """
        reg = r"^"+name+r"([0-9]+$|$)"
        def unattached_match(x):
            return re.search(reg, x) and not self.__identified_clients.get(x)["attached"]
        matching_names = list(filter(unattached_match,self.__ident2sid.keys()))
        while not len(matching_names):
            await asyncio.sleep(.1)
            matching_names = list(filter(unattached_match,self.__ident2sid.keys()))

        cl = self.__identified_clients.get(matching_names[0])
        self.__identified_clients[matching_names[0]]["attached"] = True

        # TODO Check sid
        await self.sio.emit("update_id",json.dumps({"new_id":local_id}),to=cl["sid"])
        return cl

    def start(self, task: Callable[[None], None], listeners: list[EventSlave]) -> None:
        """
            This method is blocking.

            Starts an emitting sequence and runs a tasks that embeds
            calls to `EventMaster.__instance.sio.emit()`

            The starting of the game is starting when for all `EventSlave` instances in `listeners`, the `.listen()` future is fulfilled.

            If `len(listeners)==0` the EventMaster emits events
            in the void.

            Args:
                task (Callable[[None],None]): task calling `EventMaster.sio.emit()`
        """
        slaves = list(filter(lambda x:isinstance(x,EventSlave),listeners))
        self.expected_clients = len(slaves)

        # Sets the runner up and starts the tcp server
        self.event_loop.run_until_complete(self.runner.setup())
        #print(self.port)
        site = web.TCPSite(self.runner, self.hostname, self.port)
        self.event_loop.run_until_complete(site.start())

        async def stop():

            # Waiting for all
            logger.info(f"Waiting for listeners {self.__n_clients_connected} out of {self.expected_clients} are connected.")
            for x in slaves:
                await x.listen(master_address=f"http://{self.hostname}:{self.port!s}", keep_alive=False)

            logger.info("Starting match")

            # Launching the task
            task_future = asyncio.wrap_future(self.sio.start_background_task(task))
            await task_future

            # Cleaning up and closing the runner upon completion
            try:
                await asyncio.wait_for(self.runner.cleanup(), timeout=1)
            except asyncio.exceptions.TimeoutError:
                pass

        # Blocking call to the procedure
        self.event_loop.run_until_complete(stop())



import asyncio
import csv
from sys import platform

from split import chop


class ExecMulti:
    """
    A class to execute multiple rounds and matches of a game.

    Attributes:
        main_file (str): The main file to execute.
        num_player (int): The number of players in each match.
    """

    def __init__(self, main_file: str, log_file: str = "log.txt") -> None:
        """
        Initializes a new instance of the ExecMulti class.

        Args:
            main_file (str): The main file to execute.
        """
        self.main_file = main_file
        self.num_player = 2
        self.log_file = log_file

    async def run_round(self, folder_players: str, name_player1: str, name_player2: str, port: int, recordjs: bool):
        """
        Runs a single round of the game.

        Args:
            folder_players (str): The folder containing the player files.
            name_player1 (str): The name of player 1.
            name_player2 (str): The name of player 2.
            port (int): The port number for communication.

        Returns:
            None
        """
        if platform == "win32" :
            cmd = "python " + self.main_file + ".py" + " -t local"
            if recordjs :
                cmd += " --record"
            cmd += " -p " + str(port) + " " + folder_players+name_player1 + " " + folder_players+name_player2
        else :
            cmd = "python3 " + self.main_file + ".py" + " -t local"
            if recordjs :
                cmd += " --record"
            cmd += " -p " + str(port) + " " + folder_players+name_player1 + " " + folder_players+name_player2
        process = await asyncio.create_subprocess_shell(cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        stdout, stderr = await process.communicate()
        with open(self.log_file,"a+") as f :
            f.write(stderr.decode(encoding="utf-8"))
        f.close()

    async def run_multiple_rounds(self, rounds: int, nb_process: int, swap: bool, folder_players: str, name_player1: str, name_player2: str, port: int = 8080, *, recordjs: bool = False):
        """
        Runs multiple rounds of the game.

        Args:
            rounds (int): The number of rounds to run.
            nb_process (int): The number of processes to use.
            swap (bool): Whether to swap the players in alternate rounds.
            folder_players (str): The folder containing the player files.
            name_player1 (str): The name of player 1.
            name_player2 (str): The name of player 2.
            port (int): The port number for communication. Default is 8080.

        Returns:
            None
        """
        for chunk in list(chop(nb_process, range(rounds))):
            list_jobs_routines = []
            for add, _ in enumerate(chunk):
                if swap:
                    if add % 2 == 0:
                        p1 = name_player1
                        p2 = name_player2
                    else:
                        p1 = name_player2
                        p2 = name_player1
                else:
                    p1 = name_player1
                    p2 = name_player2
                list_jobs_routines.append(asyncio.create_task(self.run_round(folder_players, p1, p2, port+add, recordjs)))
            await asyncio.gather(*list_jobs_routines)

    async def run_match(self, rounds_by_match: int, swap: bool, folder_players: str, name_player1: str, name_player2: str, port: int, recordjs: bool):
        """
        Runs a single match of the game.

        Args:
            rounds_by_match (int): The number of rounds per match.
            swap (bool): Whether to swap the players in alternate matches.
            folder_players (str): The folder containing the player files.
            name_player1 (str): The name of player 1.
            name_player2 (str): The name of player 2.
            port (int): The port number for communication.

        Returns:
            None
        """
        await self.run_multiple_rounds(rounds=rounds_by_match, nb_process=1, swap=swap, folder_players=folder_players, name_player1=name_player1, name_player2=name_player2, port=port, recordjs=recordjs)

    async def run_multiple_matches(self, rounds_by_match: int, nb_process: int, swap: bool, folder_players: str, csv_file: str, sep: str = ",", port: int = 8080, *, recordjs: bool = False):
        """
        Runs multiple matches of the game.

        Args:
            rounds_by_match (int): The number of rounds per match.
            nb_process (int): The number of processes to use.
            swap (bool): Whether to swap the players in alternate matches.
            folder_players (str): The folder containing the player files.
            csv_file (str): The CSV file containing the names of the players.
            sep (str): The delimiter used in the CSV file. Default is ",".
            port (int): The starting port number for communication. Default is 8080.

        Returns:
            None
        """
        with open(csv_file) as csvfile:
            spamreader = csv.reader(csvfile, delimiter=sep)
            match_table = []
            for line in spamreader:
                counter = 0
                match = []
                for name in line:
                    match.append(name)
                    if len(match) == self.num_player:
                        match_table.append(match)
                        match = []
                        counter = 0
                    else:
                        counter += 1
            for chunk in list(chop(nb_process, match_table)):
                list_jobs_routines = []
                for add, match in enumerate(chunk):
                    list_jobs_routines.append(asyncio.create_task(self.run_match(rounds_by_match, swap, folder_players, match[0], match[1], port+add, recordjs)))
                await asyncio.gather(*list_jobs_routines)



from abc import abstractmethod
from collections.abc import Iterable

from seahorse.utils.custom_exceptions import MethodNotImplementedError


class Serializable:

    @abstractmethod
    def to_json(self)->dict:
        raise MethodNotImplementedError()

    @classmethod
    @abstractmethod
    def from_json(cls,data,**kwargs)->"Serializable":
        raise MethodNotImplementedError()

    @classmethod
    def sub_serialize(cls):
        def method(x):
            if isinstance(x,Serializable):
                return x.to_json()
            elif isinstance(x,Iterable):
                return [method(w) for w in x]
            elif isinstance(x,dict):
                return {str(i):j for i,j in x.items()}
            else:
                return x.__dict__
        return method



"""Generate the code reference pages."""

from pathlib import Path

import mkdocs_gen_files
import re

for path in sorted(Path("src").rglob("*.py")):  # 
    if len(re.findall('(^|/)__',str(path))):
        continue

    module_path = path.relative_to("src").with_suffix("")  # 
    doc_path = path.relative_to("src").with_suffix(".md")  # 
    full_doc_path = Path("reference", doc_path)  # 

    parts = list(module_path.parts)

    if parts[-1] == "__init__":  # 
        parts = parts[:-1]
    elif parts[-1] == "__main__":
        continue

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:  # 
        identifier = ".".join(parts)  # 
        print("::: " + identifier, file=fd)  # 

    mkdocs_gen_files.set_edit_path(full_doc_path, path)  # 



import time
import unittest

from seahorse.game.time_manager import TimeMixin, timed_function
from seahorse.utils.custom_exceptions import (
    AlreadyRunningError,
    NotRunningError,
    SeahorseTimeoutError,
    TimerNotInitializedError,
)


class DummyClass(TimeMixin):
    def __init__(self):
        self.dummy_attr = "bob"

    @timed_function
    def only_before_timeout(self):
        return True


class MixinTestCase(unittest.TestCase):

    def setUp(self):
        self.dummy = DummyClass()

    def test_time_mixin_init_object(self):
        assert self.dummy.dummy_attr == "bob"

    def test_timer_not_init(self):
        self.assertRaises(TimerNotInitializedError, self.dummy.get_time_limit)
        self.assertRaises(TimerNotInitializedError, self.dummy.get_remaining_time)
        self.assertRaises(TimerNotInitializedError, self.dummy.start_timer)
        self.assertRaises(TimerNotInitializedError, self.dummy.stop_timer)
        self.assertRaises(TimerNotInitializedError, self.dummy.is_locked)
        self.assertRaises(TimerNotInitializedError, self.dummy.is_running)
        self.assertRaises(TimerNotInitializedError, self.dummy.get_last_timestamp)
    
    def test_time_mixin_init_timer(self):
        self.dummy.init_timer(10)
        assert self.dummy.get_time_limit() == 10
        assert self.dummy.get_remaining_time() == 10
        assert not self.dummy.is_locked()
        assert not self.dummy.is_running()

    def test_time_mixin_start_twice(self):
        self.dummy.init_timer(10)
        self.dummy.start_timer()
        self.assertRaises(AlreadyRunningError, self.dummy.start_timer)

    def test_time_mixin_stop_twice(self):
        self.dummy.init_timer(10)
        self.dummy.start_timer()
        self.dummy.stop_timer()
        self.assertRaises(NotRunningError, self.dummy.stop_timer)

    def test_time_lock(self):

        def change_attr():
            self.dummy.dummy_attr = "bob"

        def call_blocked_method():
            return self.dummy.only_before_timeout()

        self.dummy.init_timer(.5)
        self.dummy.start_timer()

        time.sleep(.1)
        self.dummy.dummy_attr = "marcel"
        assert call_blocked_method()
        time.sleep(.4)

        self.assertRaises(SeahorseTimeoutError, change_attr)
        self.assertRaises(SeahorseTimeoutError, call_blocked_method)
        assert self.dummy.is_locked()

    def tearDown(self) -> None:
        self.dummy = None
